import numpy as np
import cv2
import dlib
from scipy.spatial import Delaunay
from PIL import Image, ImageFont, ImageDraw, ImageFilter
from moviepy.editor import *
from pydub import AudioSegment
import os
# 导入自定义模块
from light_face_morph.light_util import uuid, now_time
from light_face_morph.light_logger import initLogger

# 初始化日志对象
logger = initLogger()


def get_points(image):
    """
    用 dlib 获取面部特征点
    """
    face_detector = dlib.get_frontal_face_detector()
    face_pose_predictor = dlib.shape_predictor(predictor_model)
    try:
        detected_face = face_detector(image, 1)[0]
    except:
        # print('No face detected in image {}'.format(image))
        return np.array([])
    pose_landmarks = face_pose_predictor(image, detected_face)
    points = []
    for p in pose_landmarks.parts():
        points.append([p.x, p.y])

    # 加入四个顶点和四条边的中点
    x = image.shape[1] - 1
    y = image.shape[0] - 1
    points.append([0, 0])
    points.append([x // 2, 0])
    points.append([x, 0])
    points.append([x, y // 2])
    points.append([x, y])
    points.append([x // 2, y])
    points.append([0, y])
    points.append([0, y // 2])

    return np.array(points)


def get_triangles(points):
    """
    在特征点上使用 Delaunay 三角剖分
    """
    return Delaunay(points).simplices


def affine_transform(input_image, input_triangle, output_triangle, size):
    """
    仿射变换
    """
    warp_matrix = cv2.getAffineTransform(
        np.float32(input_triangle), np.float32(output_triangle))
    output_image = cv2.warpAffine(input_image, warp_matrix, (size[0], size[1]), None,
                                  flags=cv2.INTER_LINEAR, borderMode=cv2.BORDER_REFLECT_101)
    return output_image


def morph_triangle(img1, img2, img, tri1, tri2, tri, alpha):
    """
    三角形变形，Alpha 混合
    """
    # 计算三角形的边界框
    rect1 = cv2.boundingRect(np.float32([tri1]))
    rect2 = cv2.boundingRect(np.float32([tri2]))
    rect = cv2.boundingRect(np.float32([tri]))

    tri_rect1 = []
    tri_rect2 = []
    tri_rect_warped = []

    for i in range(0, 3):
        tri_rect_warped.append(
            ((tri[i][0] - rect[0]), (tri[i][1] - rect[1])))
        tri_rect1.append(
            ((tri1[i][0] - rect1[0]), (tri1[i][1] - rect1[1])))
        tri_rect2.append(
            ((tri2[i][0] - rect2[0]), (tri2[i][1] - rect2[1])))

    # 在边界框内进行仿射变换
    img1_rect = img1[rect1[1]:rect1[1] + rect1[3], rect1[0]:rect1[0] + rect1[2]]
    img2_rect = img2[rect2[1]:rect2[1] + rect2[3], rect2[0]:rect2[0] + rect2[2]]

    size = (rect[2], rect[3])
    warped_img1 = affine_transform(img1_rect, tri_rect1, tri_rect_warped, size)
    warped_img2 = affine_transform(img2_rect, tri_rect2, tri_rect_warped, size)

    # 加权求和
    img_rect = (1.0 - alpha) * warped_img1 + alpha * warped_img2

    # 生成蒙版
    mask = np.zeros((rect[3], rect[2], 3), dtype=np.float32)
    cv2.fillConvexPoly(mask, np.int32(tri_rect_warped), (1.0, 1.0, 1.0), 16, 0)

    # 应用蒙版
    tmp = img[rect[1]:rect[1] + rect[3], rect[0]:rect[0] + rect[2]]
    img[rect[1]:rect[1] + rect[3], rect[0]:rect[0] + rect[2]] = tmp * (1 - mask) + img_rect * mask


def morph_faces(video_writer, file1, file2):
    """
    融合图片
    """
    img1 = cv2.imread(file1)
    img2 = cv2.imread(file2)

    points1 = get_points(img1)
    points2 = get_points(img2)

    # for i in range(0, 6):
    #     video_writer.write(np.uint8(img1))

    if len(points1) == 0 or len(points2) == 0:
        logger.info('无法识别人脸变换')
        for m in range(0, 18):
            alpha = float(m) / 18
            alpha2 = 1 - alpha
            dst = cv2.addWeighted(img1, alpha2, img2, alpha, 0)

            # dst = cv2.resize(dst, (img_x, img_y))
            dst = put_text(dst, False)
            video_writer.write(dst)

        for e in range(0, 12):
            img2 = put_text(img2, False)
            video_writer.write(img2)
    else:
        logger.info('识别人脸变换')
        for t in range(1, 18):
            alpha = float(t) / 18
            points = (1 - alpha) * np.array(points1) + alpha * np.array(points2)

            img1 = np.float32(img1)
            img2 = np.float32(img2)
            img_morphed = np.zeros(img1.shape, dtype=img1.dtype)

            triangles = get_triangles(points)
            for ti in triangles:
                x = ti[0]
                y = ti[1]
                z = ti[2]

                tri1 = [points1[x], points1[y], points1[z]]
                tri2 = [points2[x], points2[y], points2[z]]
                tri = [points[x], points[y], points[z]]
                morph_triangle(img1, img2, img_morphed, tri1, tri2, tri, alpha)

            img_morphed = put_text(img_morphed, True)
            video_writer.write(np.uint8(img_morphed))

        for e in range(0, 12):
            img2 = put_text(img2, True)
            video_writer.write(np.uint8(img2))


def handle_pic(file):
    # im = Image.new("RGB", (img_x, img_y), "white")
    # im_nd_array = np.array(im)

    img = Image.open(file, "r")
    image = cv2.cvtColor(np.asarray(img), cv2.COLOR_RGB2BGR)
    # 双三次插值
    height, width = image.shape[:2]  # 获取原图像的水平方向尺寸和垂直方向尺寸。

    s = height / width

    # if height <= img_y and width <= img_x:
    #     return

    if height > img_y:
        height = img_y
        width = int(height / s)
    if width > img_x:
        width = img_x
        height = int(width * s)

    res = cv2.resize(image, (width, height), interpolation=cv2.INTER_AREA)

    # 创建滤波器，使用不同的卷积核
    img_e = Image.fromarray(cv2.cvtColor(res, cv2.COLOR_BGR2RGB))
    gary2 = img_e.filter(ImageFilter.DETAIL)
    # #图像点运算
    # gary3 = gary2.point(lambda i: i * 0.9)
    img_convert_nd_array = cv2.cvtColor(np.asarray(gary2), cv2.COLOR_RGB2BGR)
    height1, width1 = img_convert_nd_array.shape[:2]
    temp_h = int((img_y - height1) / 2)
    temp_w = int((img_x - width1) / 2)
    a = cv2.copyMakeBorder(img_convert_nd_array, temp_h, img_y - temp_h - height1, temp_w, img_x - temp_w - width1,
                           cv2.BORDER_CONSTANT, value=[255, 255, 255])
    cv2.imencode('.png', a)[1].tofile(file)  # 保存图片


def put_text(img, np_flag):
    # 中文水印
    if np_flag:
        img = np.uint8(img)
    pil_image = Image.fromarray(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
    font = ImageFont.truetype(ttc_file, 20)
    color = (255, 140, 0)
    pos = (405, 15)
    draw = ImageDraw.Draw(pil_image)
    draw.text(pos, watermark_text, font=font, fill=color)

    cv_img = cv2.cvtColor(np.asarray(pil_image), cv2.COLOR_RGB2BGR)
    return cv_img
    # 英文水印
    # cv2.putText(img, 'adamia', (405, 15), cv2.FONT_HERSHEY_SIMPLEX, 0.8, (23, 76, 45), 2)
    # return img


def get_audio(music_name, video_len, out_file):
    music = AudioSegment.from_wav(music_name)
    music_time = len(music)
    if video_len * 1000 <= music_time:
        clip = music[:video_len * 1000]
        clip.export(out_file, format='wav')
    else:
        music_result = music
        for i in range(int(video_len * 1000 / music_time)):
            music_result = music_result + music
        music_result.export(out_file, format="wav")
        music_result_time = len(music_result)
        if video_len * 1000 <= music_result_time:
            clip = music_result[:video_len * 1000]
            clip.export(out_file, format='wav')


# 定义业务常量
predictor_model = 'local_resource/shape_predictor_68_face_landmarks.dat'
ttc_file = 'local_resource/msyh.ttc'
watermark_text = '阿达米亚出品'
img_x = 540
img_y = 960


def handle(content_data):
    # 构建本次流水号
    str_uid = uuid()

    if not os.path.exists('cache_data'):
        os.mkdir('cache_data')

    mp4_file = 'cache_data/' + str_uid + '.mp4'
    wav_file = 'cache_data/' + str_uid + '.wav'

    video_writer = cv2.VideoWriter(filename=mp4_file, fourcc=cv2.VideoWriter_fourcc('m', 'p', '4', 'v'), fps=8,
                                   frameSize=(img_x, img_y))

    logger.info('开始执行')
    # 参数为本地文件
    starts = ['local_resource/00.jpg', 'local_resource/01.jpg']
    ends = ['local_resource/02.jpg']

    try:
        for i in starts:
            image_0 = cv2.imread(i)
            for j in range(0, 6):
                video_writer.write(image_0)

        count = 0
        last = ''
        for i in content_data:
            handle_pic(i)
            if count == 0:  # this is fake
                last = i
                for j in range(0, 6):
                    img1 = cv2.imread(i)
                    img1 = put_text(img1, False)
                    video_writer.write(np.uint8(img1))
            else:
                now = i
                if len(last) != 0 and len(now) != 0:
                    morph_faces(video_writer, last, now)
                last = now
            count = count + 1

        for i in ends:
            image_0 = cv2.imread(i)
            for j in range(0, 6):
                image_0 = put_text(image_0, False)
                video_writer.write(image_0)
        # 释放资源
        video_writer.release()

        logger.info('读取视频和音频 并截取音频长度为视频长度')
        # 读取视频
        video = VideoFileClip(mp4_file)
        print(video.duration)
        # 读取音频
        get_audio('local_resource/42.wav', video.duration, wav_file)

        logger.info('合成视频和音频')
        audio = AudioFileClip(wav_file)
        # 设置视频的音频
        video = video.set_audio(audio)
        # 保存新的视频文件
        if not os.path.exists('result_data'):
            os.mkdir('result_data')

        mp4_result_file = now_time() + '_' + str_uid + '.mp4'
        mp4_result = 'result_data/' + mp4_result_file

        video.write_videofile(mp4_result)
    finally:
        logger.info('释放视频缓存')

        if video_writer.isOpened():
            video_writer.release()

        logger.info('移除临时文件')
        if os.path.exists(wav_file):
            os.remove(wav_file)
        if os.path.exists(mp4_file):
            os.remove(mp4_file)

    return mp4_result_file
    logger.info('结束执行')


# 主方法
if __name__ == '__main__':
    handle_data = ['source_data/z1.jpg', 'source_data/3.jpg', 'source_data/4.jpg', 'source_data/5.jpg',
                   'source_data/1.jpg']
    handle(handle_data)
